CON

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000


VAR
        long          t[4]                              ' Array of temperature readings
        long          index                             ' Current index into t[4]
        long          compressor                        ' Current compressor state
        long          total                             ' Hack variable to enable periodic display updates
                                                        ' on change only
        long          average                           ' Running accumulator of temp samples
        long          ctoggle                           ' Last time compressor changed state
        long          overridetime                      ' Compressor override counter
        long          lowval                            ' Value of lowest temp reached for stable time
        long          lowtimer                          ' Min time of lowest temp reached
        long          curavg                            ' Current average temp
        
        long          sfirstrun                         ' Need to setup sample array
        long          sampcount                         ' Number of samples to average
        long          samples                           ' Pointer to average array
        long          loops
        word          sarray[512]                       ' Array for moving average smoothing

        byte          inline[128]                       ' UART input buffer
        byte          inpos                             ' inline pointer
        byte          pbuf[128]                         ' Parser buffer
        byte          ppos                              ' pbuf pointer
        byte          state                             ' Thermostat state machine
        byte          display                           ' display periodic data?
        
' Actual setpoint variables

        long          minoffsecs                        ' Compressor minimum off time
        long          minonsecs                         ' Compressor minimum run time
        long          maxoffsecs                        ' Compressor maximum off time                        
        long          maxonsecs                         ' Compressor maximum run time                         
        long          hightrip                          ' High temperature trip point
        long          lowtrip                           ' Low temperature trip point
        long          lowtime                           ' Time at stable minimum before switchoff

OBJ

  ow : "One_Wire_001"
  text : "USB_Serial_003"
'  text : "hwuart_001"
  
PUB Main
  index := 1
  state := 0
  display := 1
  minoffsecs := 600
  maxoffsecs := 1800
  minonsecs := 900
  maxonsecs := 4200
  hightrip := 265
  lowtrip :=  215
  sampcount := 100
  sfirstrun := 1
  lowtime := 1200
   
  ow.Start_temp
  text.start
  loops := 0
  ow.Start_Conversion(24)
  set_compressor(0)
  repeat
    repeat until total <> (t[1]+t[2]+t[3]+compressor)
      loops++
      poll_serial
      update_temperature
      check_compressor
      case state

        0:  if ow.get_secs < 180                        ' power up delay
              next
            text.str(string("Powerup delay ended",10,13))
            send_data
            set_compressor(1)
            state := 2
            next

        1:  if ow.get_secs < (ctoggle+minoffsecs) ' not off for long enough
              next

            if curavg > hightrip                         ' top trip point
              text.str(string("High temp trip point reached",10,13))
              send_data
              set_compressor(1)
              state := 2
              next

            if (ctoggle+maxoffsecs) < ow.get_secs       ' off for too long
              text.str(string("Compressor off for too long",10,13))
              send_data
              set_compressor(1)
              state := 2
            next
                        
        2:  if curavg < lowval                          ' If temp dropping still then reset
              lowval := curavg
              lowtimer := ow.get_secs

            ' This allows for some jitter. Must not go higher than lowest reached + 0.2
            if curavg > (lowval+2)                                          
              lowtimer := ow.get_secs
              lowval := curavg
              
            if (ctoggle+minonsecs) > ow.get_secs ' not on long enough
              next

            if (lowtimer + lowtime) < ow.get_secs
              text.str(string("Low temperature stable for "))
              text.dec(lowtime)
              text.str(string(" Seconds",10,13))
              set_compressor(0)
              state := 1
              next

            if curavg < lowtrip            ' bottom trip point
              text.str(string("Low temp trip point reached",10,13))
              send_data
              set_compressor(0)
              state := 1
              next

            if (ctoggle+maxonsecs) < ow.get_secs        ' on too long
              text.str(string("Compressor on for too long",10,13))
              send_data
              set_compressor(0)
              state := 1
            next

        3:  if (ctoggle+overridetime) > ow.get_secs     ' manual override
              next
            if compressor == 0
              set_compressor(1)
              state := 2
              next
            if compressor == 1
              set_compressor(0)
              state := 1
              next  
    if display == 1
      send_data
    total := t[1]+t[2]+t[3]+compressor


PRI do_set(line) | p1, p2, p3
 p1 := get_decimal(line)
 p2 := get_decimal(line)
 
 case p1
  0 : text.str(string("Set parameters :"))
      text.str(string(10,13))
      text.str(string("1: Low setpoint                  (180-400)   current "))
      text.dec(lowtrip)
      text.str(string(10,13))
      text.str(string("2: High setpoint                 (220-400)   current "))
      text.dec(hightrip)
      text.str(string(10,13))
      text.str(string("3: Min off time                  (120-1800)  current "))
      text.dec(minoffsecs)
      text.str(string(10,13))
      text.str(string("4: Max off time                  (300-3600)  current "))
      text.dec(maxoffsecs)
      text.str(string(10,13))
      text.str(string("5: Min on time                   (300-3600)  current "))
      text.dec(minonsecs)
      text.str(string(10,13))
      text.str(string("6: Max on time                   (1200-5400) current "))
      text.dec(maxonsecs)
      text.str(string(10,13))
      text.str(string("7: Smoothing sample counter      (1-512)     current "))
      text.dec(sampcount)
      text.str(string(10,13))
      text.str(string("8: Reset duty cycle counters                 current "))
      text.dec(ow.get_secon)
      text.str(string(" - "))
      text.dec(ow.get_secoff)
      text.str(string(10,13))
      text.str(string("9: Min lowest stable temp timer  (600-1800)  current "))
      text.dec(lowtime)
      text.str(string(10,13))
      return
      
  1 : if p2 < 180 or p2 > 400
        text.str(string("Low setpoint out of range (180-400)",10,13))
        return
      lowtrip := p2
      text.str(string("Low setpoint set to "))
      text.dec(lowtrip)
      text.str(string(10,13))
      return

  2 : if p2 < 220 or p2 > 400
        text.out(string("High setpoint out of range (220-400)",10,13))
        return
      hightrip := p2
      text.str(string("High setpoint set to "))
      text.dec(hightrip)
      text.str(string(10,13))
      return

  3 : if p2 < 120 or p2 > 1800
        text.str(string("Min off time out of range (120-1800)",10,13))
        return
      minoffsecs := p2
      text.str(string("Minimum off time set to "))
      text.dec(minoffsecs)
      text.str(string(" seconds",10,13))
      return

  4 : if p2 < 300 or p2 > 3600
        text.str(string("Max off time out of range (300-3600)",10,13))
        return
      maxoffsecs := p2
      text.str(string("Maximum off time set to "))
      text.dec(maxoffsecs)
      text.str(string(" seconds",10,13))
      return

  5 : if p2 < 300 or p2 > 3600
        text.str(string("Min on time out of range (300-3600)",10,13))
        return
      minonsecs := p2
      text.str(string("Minimum on time set to "))
      text.dec(minonsecs)
      text.str(string(" seconds",10,13))
      return

  6 : if p2 < 1200 or p2 > 5400
        text.str(string("Max on time out of range (1200-3600)",10,13))
        return
      maxonsecs := p2
      text.str(string("Maximum on time set to "))
      text.dec(maxonsecs)
      text.str(string(" seconds",10,13))
      return

  7 : if p2 < 1 or p2 > 512
        text.str(string("Sample count out of range (1-512)",10,13))
        return
      sampcount := p2
      text.str(string("Smoothing sample count set to "))
      text.dec(sampcount)
      text.str(string(" seconds",10,13))
      samples := 0
      average := 0
      sfirstrun := 1
      return
  8 : ow.reset_timers
      text.str(string("Duty cycle counters reset",10,13))
      return
  
  9 : if p2 < 600 or p2 > 1800
        text.str(string("Min low stable temp timer out of range (600-1800)",10,13))
        return
      lowtime := p2
      text.str(string("Minimum low stable temp timer set to "))
      text.dec(lowtime)
      text.str(string(" seconds",10,13))
      return
  
PRI tempout(val)
  text.dec(val/2)
  text.out(".")
  if (val & 1) > 0
    text.out("5")
  else
    text.out("0")

PRI Get_Char(line) : val
  val := byte[line][ppos]
  ppos += 1
  
PRI Get_Decimal(line) : val | start, end, i, j, k, m
  j := 0
  i := ppos
  repeat while (byte[line][i] < 48) or (byte[line][i] > 57)
    if byte[line][i] == 0
      return 0
    i += 1
  start := i
  repeat while (byte[line][i] => 48) and (byte[line][i] =< 57)
    end := i
    i += 1
  i := end
  k := 1
  repeat while i => start
    j := j + ((byte[line][i]-48) * k)
    i -= 1
    k := k * 10
  val := j
  ppos := end+1

PRI parse_line(line) | i, j, k, p0, p1
  ppos := 0
  i := get_char(line)
  case i
    "d" : if get_decimal(line) == 0
            text.str(string("Periodic display disabled",10,13))
            display := 0
          else
            text.str(string("Periodic display enabled",10,13))
            display := 1
          send_data
          return  
    "e" : text.dec(get_decimal(line))
          return
    "f" : text.dec(get_decimal(line))
          text.out("!")
          text.dec(get_decimal(line))
          return
    "c" : p0 := get_decimal(line)
          p1 := get_decimal(line)
          set_compressor_time(p0,p1)
          return 
    "s" : do_set(line)
          return
    "l" : text.dec(loops/ow.get_secs)
          text.str(string(" mainline loops per second",10,13))
          return
    "a" : i := 0
          j := 0
          k := 0
          repeat
            if i == samples
              text.out("x")
            k := k + sarray[i]
            text.dec(sarray[i])
            text.out(",")
            i++
            j++
            if j > 9
              j := 0
              text.str(string(10,13))
          until i => sampcount
          text.dec(k / sampcount)
          text.out(",")
          text.dec(k)
          text.out(",")
          text.dec(average)
          text.out(",")
          text.dec(average-k)          
          text.str(string(10,13))
          return
  text.str(string("Help Menu",10,13,"s - set parameters",10,13))
  text.str(string("c x y - Override compressor to state (x) for (y) seconds",10,13))
  text.str(string("d x - Set periodic dump of data on(1) / off(0) and display data",10,13))
  text.str(string("a - Dump average sample array",10,13))
  
  return
          
          
PRI poll_serial | i, h
  i := text.rxcheck
  if i > -1
    h := 0
    if i == 8                                           ' Manage backspace
      text.out(i)
      text.out(" ")
      text.out(i)
      h := 1
      if inpos > 0
        inpos -= 1
      return
    if i == 13                                      ' Manage <enter>
      text.out(i)
      text.out(10)
      h := 1
      inline[inpos] := 10
      inpos += 1
      inline[inpos] := 13
      inpos += 1
      inline[inpos] := 0
      i := 0
      repeat while (i < 128 and inline[i] <> 0)
        pbuf[i] := inline[i]
        i += 1
      inpos := 0
      pbuf[i] := 0
      parse_line(@pbuf)
      return
    ' Need to add a timeout here also!
    text.out(i)                                         ' Default echo to console
    inline[inpos] := i
    inpos += 1
    return
         
PRI update_temperature | temp, i, j, k
  if ow.Get_Temperature <> 0
    temp := ow.Get_Temperature & $1FF
    k := (temp*10)/2
    if index == 3
      if sfirstrun == 1
        sfirstrun := 0
        i := 0
        repeat
          sarray[i] := k
          i++
        until i > 512
        average := k * sampcount
        samples := 0
      else
        sarray[samples] := k
        samples++
        if samples > sampcount
          samples := 0
        i := sarray[samples]
        average := average + k
        average := average - i
      curavg := average/sampcount
    t[index] := temp
    index++
    if index > 3
      index := 1
    ow.Start_Conversion(index+23)

PRI check_compressor
  if ina[18] == 0
    compressor := 1
  else
    compressor := 0

PRI send_data | sec, tmin, hour, d1, d2, d3, tempval
  if inpos <> 0
    return
  tempval := ow.Get_Time
  sec := (tempval) & $FF
  tmin := (tempval >> 8) & $FF
  hour := (tempval >> 16) & $FF
  if hour < 10
    text.out("0")
  text.dec(hour)
  text.out(":")
  if tmin < 10
    text.out("0")
  text.dec(tmin)
  text.out(":")
  if sec < 10
    text.out("0")
  text.dec(sec)
  text.out(",")
  tempout(t[1])
  text.out(",")
  tempout(t[2])
  text.out(",")
  tempout(t[3])
  text.out(",")
  text.dec(compressor)
  text.out(",")
  text.dec(ow.Get_Secon)
  text.out(",")
  text.dec(ow.Get_Secoff)
  text.out(",")
  d1 := ow.Get_Secon
  d2 := ow.Get_Secoff
  d3 := (d2*1000) / ((d2+d1))
  text.dec(d3)
  text.out(",")
  text.dec((average)/sampcount)
  text.out(",")
  text.dec(ow.get_secs - ctoggle)
  if state == 2
    text.out(",")
    text.dec(ow.get_secs - lowtimer)
  text.str(string(10,13))
  

PRI set_compressor(val)
  text.str(string("Compressor was "))
  case compressor
    0: text.str(string("off for "))
    1: text.str(string("on for "))
  text.dec(ow.get_secs - ctoggle)
  text.str(string(" Seconds",10,13))
  ctoggle := ow.get_secs
  text.str(string("Compressor Switched "))
  text.dec(val)
  text.str(string(10,13))
  outa[19] := val
  dira[19] := val

PRI set_compressor_time(sta,time)
  ctoggle := ow.get_secs
  if sta == 0 and time < minoffsecs
    text.str(string("Compressor off duration too short",10,13))
    return
  text.str(string("Compressor Switched "))
  case sta
    0 : text.str(string("Off for "))
    1 : text.str(string("On for "))
  text.dec(time)
  text.str(string(" Seconds ("))
  text.dec(time/60)
  text.str(string(") Minutes",10,13))
  state := 3
  overridetime := time
  set_compressor(sta)   